package quimufu.colourful_portals.portal;

import net.minecraft.block.BlockState;
import net.minecraft.block.Blocks;
import net.minecraft.entity.Entity;
import net.minecraft.registry.RegistryKey;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.DyeColor;
import net.minecraft.util.Identifier;
import net.minecraft.util.math.BlockBox;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Direction;
import org.jetbrains.annotations.Nullable;
import quimufu.colourful_portals.config.ColourfulPortalConfig;
import quimufu.colourful_portals.general_util.LinkedList;

import java.util.List;
import java.util.Set;

import static quimufu.colourful_portals.ColourfulPortalsMod.*;
import static quimufu.colourful_portals.Components.PORTAL_CANDIDATE_LIST;

public class PortalManager {

    private final PortalListComponent portalCandidateList;
    private final PortalListComponent portalList;

    public PortalLinkingSystem getLinkingSystem() {
        return linkingSystem;
    }

    private final PortalLinkingSystem linkingSystem;

    public PortalManager(PortalLinkingSystem portalLinkingSystem, PortalListComponent portalCandidateList, PortalListComponent portalList) {
        this.portalCandidateList = portalCandidateList;
        this.portalList = portalList;
        this.linkingSystem = portalLinkingSystem;
    }

    public void onPortalBlockPlaced(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockPlaced");
        LOGGER.debug("onPortalBlockPlaced, {}", blockId);

        Identifier dimIdentifier = getDimId(world);

        //delete portalCandidates obstructed by PortalBlocks for consistency
        List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dimIdentifier);
        LOGGER.debug("got containing PortalCandidates , {}", portalCandidates);
        for (PortalRepresentation portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate.location(), blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);

                portalCandidateList.removePortal(blockId, portalCandidate);
            }
        }
        //find new portalCandidates created by PortalBlock placement
        List<BlockBox> portals = PortalHelper.findPortalCandidates(world, pos, blockId);
        LOGGER.debug("new portalCandidates found, {}", portals);

        for (BlockBox portal : portals) {
            PortalRepresentation portalRepresentationWithDim = new PortalRepresentation(portal, dimIdentifier);
            portalCandidateList.createPortal(blockId, portalRepresentationWithDim);
        }
        world.getProfiler().pop();
    }

    public void onPortalBlockBroken(ServerWorld world, BlockPos pos, Identifier blockId) {
        world.getProfiler().push("onPortalBlockBroken");
        LOGGER.debug("onPortalBlockBroken, {}", blockId);

        PortalListComponent portalCandidateList = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());

        //check portalCandidate validity
        Identifier dimensionIdentifier = getDimId(world);
        List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dimensionIdentifier);
        LOGGER.debug("getContainingPortals, {}", portalCandidates);
        for (PortalRepresentation portalCandidate : portalCandidates) {
            if (!PortalHelper.isValidCandidate(world, portalCandidate.location(), blockId)) {
                LOGGER.debug("invalid, {}", portalCandidate);
                portalCandidateList.removePortal(blockId, portalCandidate);
            }
        }

        //add portalCandidates deobstructed by PortalBlock removal
        for (Direction direction : Direction.values()) {
            List<BlockBox> newPortalCandidates = PortalHelper.findPortalCandidates(world, pos.add(direction.getVector()), blockId);

            for (BlockBox newPortal : newPortalCandidates) {
                LOGGER.debug("potentially new candidate after deobstruction , {}", newPortal);

                PortalListComponent portalListComponent = PORTAL_CANDIDATE_LIST.get(world.getLevelProperties());
                PortalRepresentation portalRepresentationWithDim =
                        new PortalRepresentation(newPortal, dimensionIdentifier);
                portalListComponent.createPortal(blockId, portalRepresentationWithDim);
            }
        }


        //check portal validity
        List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dimensionIdentifier);
        boolean changed = false;
        for (PortalRepresentation portal : portals) {
            if (!PortalHelper.isValidPortal(world, portal.location(), blockId)) {
                LOGGER.debug("portal became invalid ,{} {}", dimensionIdentifier, portal);

                portalList.removePortal(blockId, portal);
                linkingSystem.unLinkPortal(portal);
                removePortalBlocks(world, portal.location());
                changed = true;
            }

        }
        if (changed) {
            fixPortalGroup(blockId, portalList, world.getServer());
        }

        world.getProfiler().pop();
    }

    private void removePortalBlocks(ServerWorld world, BlockBox portal) {
        BlockPos.stream(portal)
                .filter(blockPos -> world.getBlockState(blockPos).isOf(PORTAL_BLOCK))
                .forEach(blockPos -> world.setBlockState(blockPos, Blocks.AIR.getDefaultState()));
    }

    private void fixPortalGroup(Identifier blockId, PortalListComponent portalList, MinecraftServer server) {
        LinkedList<PortalRepresentation> portalsWithDim = portalList.getPortals(blockId);
        if (portalsWithDim.size() == 1) {
            PortalRepresentation portalRepresentation = portalsWithDim.getFirst();
            ServerWorld world = getPortalWorld(portalRepresentation, server);
            if (world == null) {
                LOGGER.error("error fixing portalGroup, dimensionId {} was null", portalRepresentation.dimensionId());
                return;
            }
            portalList.removePortal(blockId, portalRepresentation);
            linkingSystem.unLinkPortal(portalRepresentation);
            removePortalBlocks(world, portalRepresentation.location());
            return;
        }
        for (PortalRepresentation portalRepresentation : portalsWithDim) {
            assurePortalBlocksPlaced(blockId, portalRepresentation, server);
        }
        linkingSystem.linkPortals(portalsWithDim);

    }

    private void assurePortalBlocksPlaced(Identifier blockId, PortalRepresentation portalRepresentation, MinecraftServer server) {
        DyeColor color = ColourfulPortalConfig.colorOf(blockId.toString());
        if (color != null) {
            Direction.Axis portalPlaneAxis = PortalHelper.getAxisW(portalRepresentation.location())
                    .rotateClockwise(Direction.Axis.Y)
                    .getAxis();
            ServerWorld portalWorld = getPortalWorld(portalRepresentation, server);
            if (portalWorld == null) {
                LOGGER.error("error placing portalRepresentation planes, dimensionId {} was null", portalRepresentation.dimensionId());
                return;
            }
            BlockState portalBlockState = PORTAL_BLOCK.getStateWith(color, portalPlaneAxis);

            PortalHelper.insideOf(portalRepresentation.location())
                    .forEachRemaining(blockPos -> {
                        BlockState blockState = portalWorld.getBlockState(blockPos);
                        if (blockState.isAir() || blockState.getFluidState().isOf(PORTAL_FLUID)) {
                            portalWorld.setBlockState(blockPos, portalBlockState);
                        } else if (blockState.isReplaceable()) {
                            portalWorld.breakBlock(blockPos, true);
                            portalWorld.setBlockState(blockPos, portalBlockState);
                        }
                    });
        }

    }


    @Nullable
    private ServerWorld getPortalWorld(PortalRepresentation fromPortalRepresentation, MinecraftServer server) {
        return server.getWorld(RegistryKey.of(RegistryKeys.WORLD, fromPortalRepresentation.dimensionId()));
    }

    public boolean tryIgnite(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal ignition");
        LOGGER.info("portal ignition");
        Set<Identifier> blockIds = portalCandidateList.getBlockIds();
        boolean ret = false;
        for (Identifier blockId : blockIds) {
            Identifier dim = getDimId(world);
            List<PortalRepresentation> portalCandidates = portalCandidateList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation current : portalCandidates) {
                if (PortalHelper.isValidPortal(world, current.location(), blockId)) {
                    PortalRepresentation next = portalCandidateList.getNext(blockId, current);
                    ServerWorld nextWorld = getPortalWorld(next, world.getServer());
                    if (!next.equals(current)
                        && (!portalList.containsPortal(blockId, current) || !portalList.containsPortal(blockId, next))
                        && PortalHelper.isPortalPlaceable(nextWorld, next.location(), blockId)) {
                        portalList.createPortal(blockId, current);
                        portalList.createPortal(blockId, next);
                        ret = true;
                    }
                    fixPortalGroup(blockId, portalList, world.getServer());
                }
            }
        }
        world.getProfiler().pop();
        return ret;
    }

    public boolean canExtend(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal extension check");
        Set<Identifier> blockIds = portalList.getBlockIds();
        Identifier dim = getDimId(world);

        for (Identifier blockId : blockIds) {
            List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation portal : portals) {
                if (PortalHelper.isValidPortal(world, portal.location(), blockId)) {
                    PortalRepresentation last = portalList.getLast(blockId);
                    PortalRepresentation next = portalCandidateList.getNext(blockId, last);
                    ServerWorld nextWorld = getPortalWorld(next, world.getServer());

                    if (!portalList.containsPortal(blockId, next)
                        && PortalHelper.isPortalPlaceable(nextWorld, next.location(), blockId)) {
                        world.getProfiler().pop();
                        return true;
                    }
                }
            }

        }

        world.getProfiler().pop();
        return false;
    }

    public boolean extend(ServerWorld world, BlockPos pos) {
        world.getProfiler().push("portal extension");
        Set<Identifier> blockIds = portalList.getBlockIds();
        Identifier dim = getDimId(world);

        boolean ret = false;
        for (Identifier blockId : blockIds) {
            List<PortalRepresentation> portals = portalList.getContainingPortals(blockId, pos, dim);
            for (PortalRepresentation portal : portals) {
                if (PortalHelper.isValidPortal(world, portal.location(), blockId)) {
                    PortalRepresentation last = portalList.getLast(blockId);
                    PortalRepresentation next = portalCandidateList.getNext(blockId, last);
                    ServerWorld nextWorld = getPortalWorld(next, world.getServer());

                    if (!portalList.containsPortal(blockId, next)
                        && PortalHelper.isPortalPlaceable(nextWorld, next.location(), blockId)) {
                        portalList.createPortal(blockId, next);
                        ret = true;
                    }
                    fixPortalGroup(blockId, portalList, world.getServer());
                }
            }

        }

        world.getProfiler().pop();
        return ret;
    }

    public static Identifier getDimId(ServerWorld world) {
        return world.getDimensionEntry().getKey().orElseThrow().getValue();
    }

    public void onPortalPassed(Entity entity, BlockPos pos, ServerWorld world, Direction.Axis a) {
        linkingSystem.onPortalPassed(entity, pos, world, a);
    }
}
